#!/usr/bin/env python
#

"""
Test XML parsing routines.

These tests check that we don't lose any information by converting
an object to XML then parsing it again.  The round trip should
produce an object that is identical to the one we started with.
"""

from __future__ import absolute_import

import unittest

from pywbem import tupletree, tupleparse
from pywbem import CIMInstance, CIMInstanceName, CIMClass, CIMProperty, \
    CIMParameter, CIMQualifier, Uint8, Uint16, Uint32


class TupleTest(unittest.TestCase):
    """ Top level class for testing tupleparse"""

    def _run_single(self, obj):
        """
        Execute the test on a single object.  This test converts the object to
        xml and then parses it.  The result must be the same as the original
        object.
        """
        # Convert object to xml

        xml = obj.tocimxml().toxml()

        # Parse back to an object

        result = tupleparse.parse_any(tupletree.xml_to_tupletree_sax(xml))

        # Assert that the before and after objects should be equal

        self.assertEqual(obj, result,
                         '\nbefore: %s\nafter:  %s' %
                         (xml, result.tocimxml().toxml()))


class RawXMLTest(unittest.TestCase):
    """Test parse of any XML to object"""

    def _run_single(self, xml, obj):
        """
        Parse the xml parameter to an object and compare it with the obj
        parameter
        """
        # Parse raw XML to an object

        result = tupleparse.parse_any(tupletree.xml_to_tupletree_sax(xml))

        # Assert XML parses to particular Python object

        self.assertEqual(obj, result,
                         'parsed XML: %s' % result)


class ParseCIMInstanceName(TupleTest):
    """Test parsing of CIMInstanceName objects."""

    def test_all(self):
        """Test multiple instance name constructions"""
        self._run_single(CIMInstanceName('CIM_Foo'))

        self._run_single(CIMInstanceName('CIM_Foo',
                                         {'Name': 'Foo',
                                          'Chicken': 'Ham'}))

        self._run_single(
            CIMInstanceName('CIM_Foo', {'Name': 'Foo',
                                        'Number': 42,
                                        'Boolean': False,
                                        'Ref': CIMInstanceName('CIM_Bar')}))

        self._run_single(
            CIMInstanceName('CIM_Foo', {'Name': 'Foo'},
                            namespace='root/cimv2'))

        self._run_single(
            CIMInstanceName('CIM_Foo', {'Name': 'Foo'},
                            host='woot.com',
                            namespace='root/cimv2'))


class ParseCIMInstance(TupleTest):
    """Test parsing of CIMInstance objects."""

    def test_all(self):
        """Test multiple CIMInstanceObjects"""
        self._run_single(CIMInstance('CIM_Foo'))

        self._run_single(
            CIMInstance(
                'CIM_Foo',
                {'string': 'string',
                 'uint8': Uint8(0),
                 'uint8array': [Uint8(1), Uint8(2)],
                 'ref': CIMInstanceName('CIM_Bar')}))

        self._run_single(
            CIMInstance('CIM_Foo',
                        {'InstanceID': '1234'},
                        path=CIMInstanceName('CIM_Foo',
                                             {'InstanceID': '1234'})))


# TODO 2/16 KS: Create test for more complete class.
class ParseCIMClass(TupleTest):
    """Test parsing of CIMClass objects."""

    def test_all(self):
        """test CIMClass Constructors"""
        self._run_single(CIMClass('CIM_Foo'))
        self._run_single(CIMClass('CIM_Foo', superclass='CIM_bar'))

        self._run_single(CIMClass(
            'CIM_CollectionInSystem',
            qualifiers={'ASSOCIATION':
                        CIMQualifier('ASSOCIATION', True,
                                     overridable=False),

                        'Aggregation':
                        CIMQualifier('Aggregation', True,
                                     overridable=False),
                        'Version':
                        CIMQualifier('Version', '2.6.0',
                                     tosubclass=False,
                                     translatable=False),
                        'Description':
                        CIMQualifier('Description',
                                     'CIM_CollectionInSystem is an '
                                     'association used to establish a '
                                     'parent-child relationship between a '
                                     'collection and an \'owning\' System '
                                     'such as an AdminDomain or '
                                     'ComputerSystem. A single collection '
                                     'should not have both a '
                                     'CollectionInOrganization and a '
                                     'CollectionInSystem association.',
                                     translatable=True)},
            properties={'Parent':
                        CIMProperty(
                            'Parent', None, type='reference',
                            reference_class='CIM_System',
                            qualifiers={'Key':
                                        CIMQualifier('Key', True,
                                                     overridable=False),
                                        'Aggregate':
                                        CIMQualifier('Aggregate', True,
                                                     overridable=False),
                                        'Max':
                                        CIMQualifier('Max', Uint32(1))}),
                        'Child':
                        CIMProperty(
                            'Child', None, type='reference',
                            reference_class='CIM_Collection',
                            qualifiers={'Key':
                                        CIMQualifier('Key', True,
                                                     overridable=False)}
                            )  # noqa: E123
                       }  # noqa: E124
            ))  # noqa: E123


# TODO 2/16 KS: Extend to all property data types.
class ParseCIMProperty(TupleTest):
    """Test parsing of CIMProperty objects."""

    def test_single_valued(self):
        """Single-valued properties"""
        self._run_single(CIMProperty('Spotty', 'Foot'))
        self._run_single(CIMProperty('Age', Uint16(32)))
        self._run_single(CIMProperty('Foo', '', type='string'))
        self._run_single(CIMProperty('Foo', None, type='string'))
        self._run_single(CIMProperty('Age', None, type='uint16',
                                     qualifiers={'Key': CIMQualifier('Key',
                                                                     True)}))

    def test_property_arrays(self):
        """Property arrays"""
        self._run_single(CIMProperty('Foo', ['a', 'b', 'c']))
        self._run_single(CIMProperty('Foo', None, type='string', is_array=True))
        self._run_single(CIMProperty('Foo', [Uint8(x) for x in [1, 2, 3]],
                                     qualifiers={'Key': CIMQualifier('Key',
                                                                     True)}))

    def test_reference_properties(self):
        """Test Reference properties"""
        self._run_single(CIMProperty('Foo', None, type='reference'))
        self._run_single(CIMProperty('Foo', CIMInstanceName('CIM_Foo')))
        self._run_single(CIMProperty('Foo', CIMInstanceName('CIM_Foo'),
                                     qualifiers={'Key': CIMQualifier('Key',
                                                                     True)}))

    def test_embeddedobject_properties(self):
        """Test EmbeddedObject properties"""
        inst = CIMInstance('Foo_Class',
                           {'one': Uint8(1), 'two': Uint8(2)})
        self._run_single(CIMProperty('Foo', inst))
        self._run_single(CIMProperty('Foo', [inst]))


# TODO 2/16 KS: Extend for all data types.
class ParseCIMParameter(TupleTest):
    """Test parsing of CIMParameter objects."""

    def test_all(self):

        # Single-valued parameters

        self._run_single(CIMParameter('Param', 'string'))

        self._run_single(CIMParameter('Param', 'string',
                                      qualifiers={'Key': CIMQualifier('Key',
                                                                      True)}))

        # Reference parameters

        self._run_single(CIMParameter('RefParam', 'reference'))

        self._run_single(CIMParameter('RefParam', 'reference',
                                      reference_class='CIM_Foo'))

        self._run_single(CIMParameter('RefParam', 'reference',
                                      reference_class='CIM_Foo',
                                      qualifiers={'Key': CIMQualifier('Key',
                                                                      True)}))

        # Array parameters

        self._run_single(CIMParameter('Array', 'string', is_array=True))

        self._run_single(CIMParameter('Array', 'string', is_array=True,
                                      array_size=10))

        self._run_single(CIMParameter('Array', 'string', is_array=True,
                                      array_size=10,
                                      qualifiers={'Key': CIMQualifier('Key',
                                                                      True)}))

        # Reference array parameters

        self._run_single(CIMParameter('RefArray', 'reference',
                                      is_array=True))

        self._run_single(CIMParameter('RefArray', 'reference',
                                      is_array=True,
                                      reference_class='CIM_Foo'))

        self._run_single(CIMParameter('RefArray', 'reference',
                                      is_array=True,
                                      reference_class='CIM_Foo',
                                      array_size=10))

        self._run_single(CIMParameter('RefArray', 'reference',
                                      is_array=True,
                                      reference_class='CIM_Foo',
                                      array_size=10,
                                      qualifiers={'Key': CIMQualifier('Key',
                                                                      True)}))


class ParseXMLKeyValue(RawXMLTest):

    def test_all(self):

        self._run_single(
            '<KEYVALUE VALUETYPE="numeric">1234</KEYVALUE>', 1234)

        self._run_single(
            '<KEYVALUE TYPE="uint32" VALUETYPE="numeric">1234</KEYVALUE>',
            1234)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
#
# (C) Copyright 2004 Hewlett-Packard Development Company, L.P.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# Author: Tim Potter <tpot@hp.com>
#

# The module docstring is used as the long package description:
"""
Pywbem is a WBEM client, written in pure Python. It supports Python 2 and
Python 3.

A WBEM client allows issuing operations to a WBEM server, using the CIM
operations over HTTP (CIM-XML) protocol defined in the DMTF standards
DSP0200 and DSP0201. The CIM/WBEM infrastructure is used for a wide variety of
systems management tasks supported by systems running WBEM servers.
See http://www.dmtf.org/standards/wbem for more information about WBEM.
"""

from __future__ import print_function, absolute_import

import re
import sys
import os
import shutil
import subprocess
import platform
from distutils.errors import DistutilsSetupError

import os_setup
from os_setup import shell, shell_check, import_setuptools, YumInstaller

# Workaround for Python 2.6 issue https://bugs.python.org/issue15881
# This causes this module to be referenced and prevents the premature
# unloading and subsequent garbage collection causing the error.
if sys.version_info[0:2] == (2, 6):
    try:
        # pylint: disable=unused-import
        import multiprocessing  # noqa: E402,F401
    except ImportError:
        pass

if sys.version_info[0] == 2:
    from xmlrpclib import Fault
else:
    from xmlrpc.client import Fault

_VERBOSE = True


def package_version(filename, varname):
    """Return package version string by reading `filename` and retrieving its
       module-global variable `varnam`."""
    _locals = {}
    with open(filename) as fp:
        exec(fp.read(), None, _locals)  # pylint: disable=exec-used
    return _locals[varname]


def _check_get_swig(swig_min_version, verbose):
    """Chewck if Swig is available in the PATH in the right version.
    Returns True if it needs to be installed/updated.
    """
    if verbose:
        print("Testing for availability of Swig >=%s in PATH..." %
              swig_min_version)
    get_swig = False
    rc, out, _ = shell("which swig")
    if rc != 0:
        if verbose:
            print("Swig is not available in PATH; need to get Swig")
        get_swig = True
    else:
        if verbose:
            print("Swig is available in PATH; testing its version...")
        out = shell_check("swig -version")
        m = re.search(r"^SWIG Version ([0-9\.]+)$", out, re.MULTILINE)
        if m is None:
            raise DistutilsSetupError("Cannot determine Swig version from "
                                      "output of 'swig -version':\n%s" % out)
        swig_version = m.group(1)
        if swig_version.split(".") < swig_min_version.split("."):
            if verbose:
                print("Installed Swig version is too old: %s; "
                      "need to get Swig" % swig_version)
            get_swig = True
        else:
            if verbose:
                print("Installed Swig version is sufficient: %s" %
                      swig_version)
    return get_swig


def install_swig(installer, dry_run, verbose):
    """Custom installer function for `os_setup` module.
    This function makes sure that Swig is installed, either by installing the
    corresponding OS-level package, or by downloading the source and building
    it.

    Parameters: see description of `os_setup` module.
    """

    swig_min_version = "2.0"

    get_swig = _check_get_swig(swig_min_version, verbose)

    if get_swig:

        system = installer.system
        distro = installer.distro

        swig_pkg_dict = {
            'Linux': {
                'redhat': "swig",
                'centos': "swig",
                'fedora': "swig",
                'debian': "swig",
                'ubuntu': "swig",
                'linuxmint': "swig",
                'suse': "swig",
            },
        }
        # default if no system/distro specific name is found:
        swig_pkg_name = "swig"
        if system in swig_pkg_dict:
            distro_dict = swig_pkg_dict[system]
            if distro in distro_dict:
                swig_pkg_name = distro_dict[distro]

        swig_version_reqs = [">=%s" % swig_min_version]

        installed = installer.ensure_installed(
            swig_pkg_name, swig_version_reqs, dry_run, verbose,
            ignore=True)

        if installed and _check_get_swig(swig_min_version, verbose):
            # Package was tampered with (e.g. swig command renamed)

            if verbose:
                print("Reinstalling Swig package over existing one...")

            installer.do_install(swig_pkg_name, swig_version_reqs, dry_run,
                                 reinstall=True)

        elif not installed:

            # Build Swig from its source
            swig_build_version = "2.0.12"
            swig_dir = "swig-%s" % swig_build_version
            swig_tar_file = "swig-%s.tar.gz" % swig_build_version
            swig_install_root = "/usr"

            if verbose:
                print("Installing prerequisite OS-level packages for building "
                      "Swig...")

            swig_prereq_pkg_dict = {
                'Linux': {
                    'redhat': [
                        "pcre-devel",
                    ],
                    'centos': [
                        "pcre-devel",
                    ],
                    'fedora': [
                        "pcre-devel",
                    ],
                    'debian': [
                        "libpcre3",
                        "libpcre3-dev",
                    ],
                    'ubuntu': [
                        "libpcre3",
                        "libpcre3-dev",
                    ],
                    'linuxmint': [
                        "libpcre3",
                        "libpcre3-dev",
                    ],
                    'suse': [
                        "pcre-devel",
                    ],
                },
            }
            # default if no system/distro specific name is found:
            swig_prereq_pkg_default = [
                "pcre-devel",
            ]

            swig_prereq_pkg_names = swig_prereq_pkg_default
            if system in swig_prereq_pkg_dict:
                distro_dict = swig_prereq_pkg_dict[system]
                if distro in distro_dict:
                    swig_prereq_pkg_names = distro_dict[distro]
            for swig_prereq_pkg_name in swig_prereq_pkg_names:

                installed = installer.ensure_installed(
                    swig_prereq_pkg_name, None, dry_run, verbose,
                    ignore=False)

            if dry_run:
                if verbose:
                    print("Dry-running: Building Swig version %s from "
                          "downloaded source, and installing to %s tree" %
                          (swig_build_version, swig_install_root))
            else:
                if verbose:
                    print("Building Swig version %s from "
                          "downloaded source, and installing to %s tree" %
                          (swig_build_version, swig_install_root))

                if os.path.exists(swig_dir):
                    if verbose:
                        print("Removing previously downloaded Swig directory: "
                              "%s" % swig_dir)
                    shutil.rmtree(swig_dir)

                if verbose:
                    print("Downloading Swig source archive: %s" % swig_tar_file)
                shell_check(
                    "wget -q -O %s http://sourceforge.net/projects/swig/files"
                    "/swig/%s/%s/download" %
                    (swig_tar_file, swig_dir, swig_tar_file), display=True)
                if verbose:
                    print("Unpacking Swig source archive: %s" % swig_tar_file)
                shell_check("tar -xf %s" % swig_tar_file, display=True)

                if verbose:
                    print("Configuring Swig build process for installing to "
                          "%s tree..." % swig_install_root)
                shell_check(
                    ["sh", "-c", "cd %s; ./configure --prefix=%s" %
                     (swig_dir, swig_install_root)], display=True)

                if verbose:
                    print("Building Swig...")
                shell_check(
                    ["sh", "-c", "cd %s; make swig" % swig_dir], display=True)

                if verbose:
                    print("Installing Swig to %s tree..." % swig_install_root)
                shell_check(
                    ["sh", "-c", "cd %s; sudo make install" % swig_dir],
                    display=True)

                if verbose:
                    print("Done downloading, building and installing Swig "
                          "version %s" % swig_build_version)


def build_moftab(verbose):  # pylint: disable=unused-argument
    """Generate the moftab modules.

    This function is called after the installation of the `pywbem`package
    and its dependencies.
    The modules are generated into the directory where the
    `mof_compiler` module runs from, i.e. the installation target.

    Because the generation depends on the packages `ply`, `six`, and
    `M2Crypto` the build function is invoked in a child Python process,
    which picks up the changed module search path.

    Note that ideally these dependencies would be specified in a
    `setup_requires` keyword, but at this point (setuptools v20), it does
    not work together woth `install_requires`. See setuptools issue #391:
    https://bitbucket.org/pypa/setuptools/issues/391/
    dependencies-listed-in-both-setup_requires
    """
    rc = subprocess.call([sys.executable, 'build_moftab.py'])
    if rc != 0:
        # Because this does not work on pip, the best compromise is to
        # tolerate a failure:
        print("Warning: build_moftab.py failed with rc=%s; the PyWBEM "
              "LEX/YACC table modules may be rebuilt on first use" % rc)


def main():
    """Main function of this script."""

    import_setuptools()
    from setuptools import setup
    from setuptools.command.build_py import build_py as _build_py

    class build_py(_build_py):
        # pylint: disable=invalid-name,too-few-public-methods
        """Custom command that extends the setuptools `build_py` command,
        which prepares the Python files before they are being installed.
        This command is used by `setup.py install` and `pip install`.

        We use this only to pick up the verbosity level.
        """
        def run(self):
            global _VERBOSE  # pylint: disable=global-statement
            _VERBOSE = self.verbose
            _build_py.run(self)

    py_version_mn = "%s%s" % (sys.version_info[0], sys.version_info[1])
    py_version_m = "%s" % sys.version_info[0]
    yuminst = YumInstaller()

    pkg_version = package_version("pywbem/_version.py", "__version__")

    args = {
        'name': 'pywbem',
        'author': 'Tim Potter',
        'author_email': 'tpot@hp.com',
        'maintainer': 'Andreas Maier',
        'maintainer_email': 'maiera@de.ibm.com',
        'description': 'pywbem - A WBEM client',
        'long_description': __doc__,
        'platforms': ['any'],
        'url': 'http://pywbem.github.io/pywbem/',
        'version': pkg_version,
        'license': 'LGPL version 2.1, or (at your option) any later version',
        'distclass': os_setup.OsDistribution,
        'cmdclass': {
            'build_py': build_py,
            'install_os': os_setup.install_os,
            'develop_os': os_setup.develop_os,
            'develop': os_setup.develop,
        },
        'packages': ['pywbem'],
        'package_data': {
            'pywbem': [
                'NEWS.md',
                'LICENSE.txt',
            ]
        },
        'scripts': [
            'wbemcli',
            'wbemcli.py',
            'mof_compiler',
            'wbemcli.bat',
            'mof_compiler.bat',
        ],
        'install_requires': [
            # These dependencies will be installed as a site package.
            # They are not useable by this setup script, if they are eggs
            # (because their path is added to a .pth file which is parsed only
            # at Python startup time).
            'six',
            'ply',
            # The PyYAML package contains the "yaml" Python package. yaml is
            # needed by the pywbem._recorder module.
            'PyYAML',
        ],
        'develop_requires': [
            # Wheel may not be installed in every system Python.
            # Wheel 0.30.0 removed Python 2.6 support.
            # Keep the condition for wheel consistent with the makefile.
            'wheel<0.30.0' if sys.version_info[0:2] == (2, 6) else 'wheel',
            # Python prereqs for 'develop' command. Handled by os_setup module.
            "coverage>=4.3",
            "pytest>=2.4",
            "pytest-cov",
            'testfixtures>=4.13.3',
            "Sphinx>=1.3" if sys.version_info[0:2] != (2, 6) else None,
            # Pinning GitPython to 2.0.8 max, due to its use of unittest.case
            # which is not available on Python 2.6.
            # TODO: Track resolution of GitPython issue #540:
            #       https://github.com/gitpython-developers/GitPython/issues/540
            "GitPython==2.0.8" if sys.version_info[0:2] != (2, 6) else None,
            "sphinx-git" if sys.version_info[0:2] != (2, 6) else None,
            "httpretty",
            "lxml",
            # Astroid is used by Pylint. Astroid 1.3 and above, and Pylint 1.4
            # and above no longer work with Python 2.6, and have been removed
            # from Pypi in 2/2016 after being available for some time.
            # Therefore, we cannot use Pylint under Python 2.6.
            # Also, Pylint does not support Python 3.
            "astroid" if sys.version_info[0:2] == (2, 7) else None,
            "pylint" if sys.version_info[0:2] == (2, 7) else None,
            "mock",
            'flake8',
            "pbr",  # needed by mock
            "twine",  # needed for upload to Pypi
        ],
        'install_os_requires': {
            # OS-level prereqs for 'install_os' command. Handled by os_setup
            # module.
            'Linux': {
                'redhat': [
                    "openssl-devel>=1.0.1",  # for M2Crypto installation
                    "gcc-c++>=4.4",         # for building Swig and for running
                                            #   Swig in M2Crypto install
                    install_swig,           # for running Swig in M2Crypto inst.
                    # Python*-devel provides Python.h for Swig run.
                    "python-devel" if py_version_m == "2" else \
                    "python%su-devel" % py_version_mn if \
                    yuminst.is_installed('python%su' % py_version_mn)[0] else \
                    "python%s-devel" % py_version_mn
                ],
                'centos': 'redhat',
                'fedora': 'redhat',
                'ubuntu': [
                    "libssl-dev>=1.0.1",
                    "g++>=4.4",
                    install_swig,
                    "python-dev" if py_version_m == "2" else \
                    "python%s-dev" % py_version_m,
                ],
                'debian': 'ubuntu',
                'linuxmint': 'ubuntu',
                'suse': [
                    "openssl-devel>=1.0.1",
                    "gcc-c++>=4.4",
                    install_swig,
                    "python-devel" if py_version_m == "2" else \
                    "python%s-devel" % py_version_m,
                ],
            },
            # TODO: Add support for Windows.
        },
        'develop_os_requires': {
            # OS-level prereqs for 'develop_os' command, in addition to those
            # defined in 'install_os_requires'. Handled by os_setup module.
            'Linux': {
                'redhat': [
                    "libxml2-devel",        # for installing Python lxml pkg
                    "libxml2",              # for installing xmllint command
                    "libxslt-devel",        # for installing Python lxml pkg
                    "libyaml-devel",        # for installing Python pyyaml pkg
                    "make",                 # PyWBEM has a makefile
                    "tar",                  # for distribution archive
                    "git",                  # used by GitPython
                ],
                'centos': 'redhat',
                'fedora': 'redhat',
                'debian': [
                    "libxml2-dev",
                    "libxml2-utils",
                    "libxslt1-dev",
                    "libyaml-dev",
                    "make",
                    "tar",
                    "git",
                ],
                'ubuntu': [
                    "libxml2-dev",
                    "libxml2-utils",
                    "libxslt1-dev",
                    "libyaml-dev",
                    "make",
                    "tar",
                    "git",
                ],
                'linuxmint': [
                    "libxml2-dev",
                    "libxml2-utils",
                    "libxslt1-dev",
                    "libyaml-dev",
                    "make",
                    "tar",
                    "git",
                ],
                'suse': [
                    "libxml2-devel",
                    "libxml2",
                    "libxslt-devel",
                    "libyaml-devel",
                    "make",
                    "tar",
                    "git",
                ],
            },
            # TODO: Add support for Windows. Some notes:
            # - install lxml from its binaries at:
            #   http://www.lfd.uci.edu/~gohlke/pythonlibs/#lxml
        },
        'classifiers': [
            'Development Status :: 5 - Production/Stable',
            'Intended Audience :: Developers',
            'Intended Audience :: System Administrators',
            'License :: OSI Approved :: '\
            'GNU Lesser General Public License v2 or later (LGPLv2+)',
            'Operating System :: OS Independent',
            'Programming Language :: Python :: 2',
            'Programming Language :: Python :: 2.6',
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.4',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
            'Topic :: Software Development :: Libraries :: Python Modules',
            'Topic :: System :: Systems Administration',
        ],
    }

    _ON_RTD = os.environ.get('READTHEDOCS', None) == 'True'

    if sys.version_info[0] == 2 and not _ON_RTD:
        # RTD does not have Swig so we cannot install M2Crypto.

        # The 'install_requires' processing in distutils does not tolerate
        # a None value in the list, so we need be truly conditional (instead
        # of adding an entry with None).
        if platform.system() == 'Windows':
            if platform.architecture()[0] == '64bit':
                m2crypto_req = 'M2CryptoWin64>=0.21'
            else:
                m2crypto_req = 'M2CryptoWin32>=0.21'
        else:
            m2crypto_req = 'M2Crypto>=0.24'
        args['install_requires'].append(m2crypto_req)

    # The ordereddict package is a backport of collections.OrderedDict
    # to Python 2.6. OrderedDict is needed by the GitPython package
    # since its 2.0.3 version (but only from 2.0.5 on it is used
    # correctly, and only from 2.0.6 on does it work on Python 2.6).
    # GitPython is needed by sphinx-git. OrderedDict is also needed
    # by the pywbem._recorder module.
    if sys.version_info[0:2] == (2, 6):
        args['install_requires'].append('ordereddict')

    # The following retry logic attempts to handle the frequent xmlrpc
    # errors that recently (9/2016) have shown up with Pypi.
    tries = 2
    while True:
        tries -= 1
        try:
            setup(**args)
        except Fault as exc:
            if tries > 0:
                print("Warning: Retrying setup() because %s was raised: %s" %
                      (exc.__class__.__name__, exc))
                continue
            else:
                raise
        else:
            break

    if 'install' in sys.argv or 'develop' in sys.argv:
        build_moftab(_VERBOSE)

    return 0


if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python
#

"""
Test functions in cim_operations
"""

from __future__ import print_function, absolute_import

import unittest

from pywbem.cim_operations import check_utf8_xml_chars, ParseError, \
    WBEMConnection

from pywbem._recorder import LogOperationRecorder, TestClientRecorder


#################################################################
# Test check_utf8_xml_chars function
#################################################################


# pylint: disable=invalid-name
class Test_check_utf8_xml_chars(unittest.TestCase):
    """Test various inputs to the check_utf8_xml_chars function"""

    def setUp(self):
        self.VERBOSE = False

    def _run_single(self, utf8_xml, expected_ok):
        """Run single test from test_all. Executes the check...
           function and catches exceptions
        """
        if self.VERBOSE:
            print('utf8_xml %s type %s ' % (utf8_xml, type(utf8_xml)))
        try:
            check_utf8_xml_chars(utf8_xml, "Test XML")
        except ParseError as exc:
            if self.VERBOSE:
                print("Verify manually: Input XML: %r, ParseError: %s" %
                      (utf8_xml, exc))
            self.assertFalse(expected_ok,
                             "ParseError unexpectedly raised: %s" % exc)
        else:
            self.assertTrue(expected_ok,
                            "ParseError unexpectedly not raised.")

    def test_all(self):
        """ Test cases. Each case tests a particular utf8 xml string"""

        # good cases
        self._run_single(b'<V>a</V>', True)
        self._run_single(b'<V>a\tb\nc\rd</V>', True)
        self._run_single(b'<V>a\x09b\x0Ac\x0Dd</V>', True)
        self._run_single(b'<V>a\xCD\x90b</V>', True)             # U+350
        self._run_single(b'<V>a\xE2\x80\x93b</V>', True)         # U+2013
        self._run_single(b'<V>a\xF0\x90\x84\xA2b</V>', True)     # U+10122

        # invalid XML characters
        if self.VERBOSE:
            print("From here on, the only expected exception is ParseError "
                  "for invalid XML characters...")
        self._run_single(b'<V>a\bb</V>', False)
        self._run_single(b'<V>a\x08b</V>', False)
        self._run_single(b'<V>a\x00b</V>', False)
        self._run_single(b'<V>a\x01b</V>', False)
        self._run_single(b'<V>a\x1Ab</V>', False)
        self._run_single(b'<V>a\x1Ab\x1Fc</V>', False)

        # correctly encoded but ill-formed UTF-8
        if self.VERBOSE:
            print("From here on, the only expected exception is ParseError "
                  "for ill-formed UTF-8 Byte sequences...")
        # combo of U+D800,U+DD22:
        self._run_single(b'<V>a\xED\xA0\x80\xED\xB4\xA2b</V>', False)
        # combo of U+D800,U+DD22 and combo of U+D800,U+DD23:
        # pylint: disable=line-too-long
        self._run_single(
            b'<V>a\xED\xA0\x80\xED\xB4\xA2b\xED\xA0\x80\xED\xB4\xA3</V>',
            False)

        # incorrectly encoded UTF-8
        if self.VERBOSE:
            print("From here on, the only expected exception is ParseError "
                  "for invalid UTF-8 Byte sequences...")
        # incorrect 1-byte sequence:
        self._run_single(b'<V>a\x80b</V>', False)
        # 2-byte sequence with missing second byte:
        self._run_single(b'<V>a\xC0', False)
        # 2-byte sequence with incorrect 2nd byte
        self._run_single(b'<V>a\xC0b</V>', False)
        # 4-byte sequence with incorrect 3rd byte:
        self._run_single(b'<V>a\xF1\x80abc</V>', False)
        # 4-byte sequence with incorrect 3rd byte that is an incorr. new start:
        self._run_single(b'<V>a\xF1\x80\xFFbc</V>', False)
        # 4-byte sequence with incorrect 3rd byte that is an correct new start:
        self._run_single(b'<V>a\xF1\x80\xC2\x81c</V>', False)


class test_create_connection(unittest.TestCase):
    """
    Test construction of WBEMConnection and those functions that do not
    depend on actually creating a connection
    """
    def test_connection_defaults(self):
        """Test creation of a connection with default constructor
           parameters.
        """
        conn = WBEMConnection('http://localhost')
        self.assertTrue(conn.url, 'http://localhost')
        self.assertEqual(conn.creds, None)
        self.assertEqual(conn.x509, None)
        self.assertEqual(conn.use_pull_operations, False)
        self.assertEqual(conn.stats_enabled, False)

    def test_no_recorder(self):
        """Test creation of wbem connection with specific parameters"""
        creds = ('myname', 'mypw')
        x509 = {'cert_file': 'mycertfile.crt', 'key_file': 'mykeyfile.key'}
        conn = WBEMConnection('http://localhost', creds,
                              default_namespace='root/blah',
                              x509=x509,
                              use_pull_operations=True,
                              enable_stats=True)
        self.assertEqual(conn.url, 'http://localhost')
        self.assertEqual(conn.creds, creds)
        self.assertEqual(conn.x509, x509)
        self.assertEqual(conn.stats_enabled, True)
        self.assertEqual(conn.use_pull_operations, True)

    def test_add_operation_recorder(self):
        """Test addition of an operation recorder"""
        conn = WBEMConnection('http://localhost')
        conn.add_operation_recorder(LogOperationRecorder())
        # pylint: disable=protected-access
        self.assertTrue(len(conn._operation_recorders), 1)

    def test_add_operation_recorders(self):
        """Test addition of multiple operation recorders"""
        conn = WBEMConnection('http://localhost')
        conn.add_operation_recorder(LogOperationRecorder())
        tcr_file = TestClientRecorder.open_file('blah.yaml', 'a')
        conn.add_operation_recorder(TestClientRecorder(tcr_file))
        # pylint: disable=protected-access
        self.assertTrue(len(conn._operation_recorders), 2)

    def test_add_same_twice(self):
        """ Test addition of same recorder twice"""
        conn = WBEMConnection('http://localhost')
        conn.add_operation_recorder(LogOperationRecorder())
        # pylint: disable=protected-access
        self.assertTrue(len(conn._operation_recorders), 1)
        try:
            conn.add_operation_recorder(LogOperationRecorder())
            self.fail('Expected exception')
        except ValueError:
            pass


if __name__ == '__main__':
    unittest.main()



#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

"""
The WBEM server API encapsulates certain functionality of a WBEM server for use
by a WBEM client application, such as determining the Interop namespace of the
server, or the management profiles advertised by the server.

This chapter has the following sections:

* :ref:`Example <Server Example>` - An example on how to use the API.

* :ref:`WBEMServer` - The :class:`~pywbem.WBEMServer` class serves as a general
  access point for clients to WBEM servers. It allows determining the Interop
  namespace of the server, or the advertised management profiles.

* :ref:`ValueMapping` - The :class:`~pywbem.ValueMapping` class maps
  corresponding values of the `Values` and `ValueMap` qualifiers of a CIM
  element and supports the translation of the actual value (often an integer)
  to the corresponding value of the `Values` qualifier.

.. note::

   The WBEM server API has been introduced in v0.9.0 as experimental and
   has been declared final in 0.10.0.

.. _`Server Example`:

Example
-------

The following example code displays some information about a WBEM server:

::

    from pywbem import WBEMConnection, WBEMServer, ValueMapping

    def explore_server(server_url, username, password):

        print("WBEM server URL:\\n  %s" % server_url)

        conn = WBEMConnection(server_url, (username, password))
        server = WBEMServer(conn)

        print("Brand:\\n  %s" % server.brand)
        print("Version:\\n  %s" % server.version)
        print("Interop namespace:\\n  %s" % server.interop_ns)

        print("All namespaces:")
        for ns in server.namespaces:
            print("  %s" % ns)

        print("Advertised management profiles:")
        org_vm = ValueMapping.for_property(server, server.interop_ns,
            'CIM_RegisteredProfile', 'RegisteredOrganization')
        for inst in server.profiles:
            org = org_vm.tovalues(inst['RegisteredOrganization'])
            name = inst['RegisteredName']
            vers = inst['RegisteredVersion']
            print("  %s %s Profile %s" % (org, name, vers))

Example output:

::

    WBEM server URL:
      http://0.0.0.0
    Brand:
      pegasus
    Version:
      2.12.0
    Interop namespace:
      root/PG_Interop
    All namespaces:
      root/PG_InterOp
      root/PG_Internal
      root/cimv2
      root
    Advertised management profiles:
      SNIA Indication Profile 1.1.0
      SNIA Indication Profile 1.2.0
      SNIA Software Profile 1.1.0
      SNIA Software Profile 1.2.0
      SNIA Profile Registration Profile 1.0.0
      SNIA SMI-S Profile 1.2.0
      SNIA Server Profile 1.1.0
      SNIA Server Profile 1.2.0
      DMTF Profile Registration Profile 1.0.0
      DMTF Indications Profile 1.1.0
"""

import re
import six

from .cim_constants import CIM_ERR_INVALID_NAMESPACE, CIM_ERR_INVALID_CLASS, \
    CIM_ERR_METHOD_NOT_AVAILABLE, CIM_ERR_NOT_SUPPORTED, CIM_ERR_NOT_FOUND
from .exceptions import CIMError
from .cim_obj import CIMInstanceName, NocaseDict
from .cim_types import CIMInt, type_from_name
from .cim_operations import WBEMConnection

__all__ = ['WBEMServer', 'ValueMapping']


class WBEMServer(object):
    """
    A representation of a WBEM server that serves as a general access point to
    a client.

    It supports determining the Interop namespace of the server, all namespaces,
    its brand and version, the advertised management profiles and finally
    allows to retrieve the central instances of a management profile with
    one method invocation regardless of whether the profile implementation
    chose the central or scoping class profile advertisement methodology
    (see :term:`DSP1033`).

    It also provides functions to subscribe for indications.
    """

    #: A class variable with the possible names of Interop namespaces that
    #: should be tried when determining the Interop namespace on the WBEM
    #: server.
    INTEROP_NAMESPACES = [
        'interop',
        'root/interop',
        'root/PG_Interop',
        # TODO: Clarify which OpenPegasus versions need root/PGInterOp?
    ]

    #: A class variable with the possible names of CIM classes for
    #: representing CIM namespaces, that should be tried when determining the
    #: namespaces on the WBEM server.
    NAMESPACE_CLASSNAMES = [
        'CIM_Namespace',
        '__Namespace',
    ]

    def __init__(self, conn):
        """
        Parameters:

          conn (:class:`~pywbem.WBEMConnection`):
            Connection to the WBEM server.
        """
        if not isinstance(conn, WBEMConnection):
            raise TypeError("conn argument of WBEMServer must be a "
                            "WBEMConnection object")
        self._conn = conn
        self._interop_ns = None
        self._namespaces = None
        self._namespace_classname = None
        self._brand = None
        self._version = None
        self._profiles = None

    def __repr__(self):
        """
        Return a representation of the :class:`~pywbem.WBEMServer` object
        with all attributes, that is suitable for debugging.
        """
        return "%s(url=%r, conn=%r, interop_ns=%s, namespaces=%s, " \
               "namespace_classname=%r, brand=%r, version=%r, " \
               "profiles=[... %s instances])" % \
               (self.__class__.__name__, self.url, self.conn, self.interop_ns,
                self.namespaces, self.namespace_classname, self.brand,
                self.version, len(self.profiles))

    @property
    def url(self):
        """
        The URL of the WBEM server, as a :term:`string`.
        """
        return self._conn.url

    @property
    def conn(self):
        """
        The connection to the WBEM server, as a
        :class:`~pywbem.WBEMConnection` object.
        """
        return self._conn

    @property
    def interop_ns(self):
        """
        The name of the Interop namespace of the WBEM server, as a
        :term:`string`.

        Raises:

            Exceptions raised by :class:`~pywbem.WBEMConnection`.
            CIMError: CIM_ERR_NOT_FOUND, Interop namespace could not be
              determined.
        """
        if self._interop_ns is None:
            self._determine_interop_ns()
        return self._interop_ns

    @property
    def namespace_classname(self):
        """
        The name of the CIM class that was found to represent the CIM
        namespaces of the WBEM server, as a :term:`string`.

        Raises:

            Exceptions raised by :class:`~pywbem.WBEMConnection`.
            CIMError: CIM_ERR_NOT_FOUND, Interop namespace could not be
              determined.
            CIMError: CIM_ERR_NOT_FOUND, Namespace class could not be
              determined.
        """
        if self._namespace_classname is None:
            self._determine_namespaces()
        return self._namespace_classname

    @property
    def namespaces(self):
        """
        A list with the names of all namespaces of the WBEM server, each
        list item being a :term:`string`.

        Raises:

            Exceptions raised by :class:`~pywbem.WBEMConnection`.
            CIMError: CIM_ERR_NOT_FOUND, Interop namespace could not be
              determined.
            CIMError: CIM_ERR_NOT_FOUND, Namespace class could not be
              determined.
        """
        if self._namespaces is None:
            self._determine_namespaces()
        return self._namespaces

    @property
    def brand(self):
        """
        Brand of the WBEM server, as a :term:`string`.

        The brand string will be one of the following:

        * ``"OpenPegasus"``, for OpenPegasus
        * ``"SFCB"``, for SFCB
        * First word of the value of the ElementName property of the
          CIM_ObjectManager instance, for any other WBEM servers.
        * ``"unknown"``, if the ElementName property is Null.

        Raises:

            Exceptions raised by :class:`~pywbem.WBEMConnection`.
            CIMError: CIM_ERR_NOT_FOUND, Interop namespace could not be
              determined.
            CIMError: CIM_ERR_NOT_FOUND, Unexpected number of
              CIM_ObjectManager instances.
        """
        if self._brand is None:
            self._determine_brand()
        return self._brand

    @property
    def version(self):
        """
        Version of the WBEM server, as a :term:`string`. `None`, if the version
        cannot be determined.

        Raises:

            Exceptions raised by :class:`~pywbem.WBEMConnection`.
            CIMError: CIM_ERR_NOT_FOUND, Interop namespace could not be
              determined.
            CIMError: CIM_ERR_NOT_FOUND, Unexpected number of
              CIM_ObjectManager instances.
        """
        if self._version is None:
            self._determine_brand()
        return self._version

    @property
    def profiles(self):
        """
        List of management profiles advertised by the WBEM server, each list
        item being a :class:`~pywbem.CIMInstance` object representing a
        CIM_RegisteredProfile instance.

        Raises:

            Exceptions raised by :class:`~pywbem.WBEMConnection`.
            CIMError: CIM_ERR_NOT_FOUND, Interop namespace could not be
              determined.
        """
        if self._profiles is None:
            self._determine_profiles()
        return self._profiles

    def get_selected_profiles(self, registered_org=None, registered_name=None,
                              registered_version=None):
        """
        List of management profiles advertised by the WBEM server and
        filtered by the input parameters for registered_org, registered_name,
        and registered_version parameters. Each list
        item is a :class:`~pywbem.CIMInstance` object representing a
        CIM_RegisteredProfile instance.

        Parameters:

            profile_org (:term:`string`) or None: the `RegisteredOrganization`
              to match the `RegisteredOrganization` of the profile. This
              match is case sensitive.
              If None, this parameter is ignored in the filter

            profile_name (:term:`string`) or None: the `RegisteredName`.
              This match is case sensitive
              If None, this parameter is ignored in the filter

            profile_version (:term:`string`) or None: the `RegisteredVersion`.
              This match is case sensitive.
              If None, this parameter is ignored in the filter

        Raises:

            Exceptions raised by :class:`~pywbem.WBEMConnection`.
            CIMError: CIM_ERR_NOT_FOUND, Interop namespace could not be
              determined.
            KeyError: If an instance in the list of profiles is incomplete
              and does not include the required properties.
        """

        org_vm = ValueMapping.for_property(self, self.interop_ns,
                                           'CIM_RegisteredProfile',
                                           'RegisteredOrganization')
        rtn = []
        for inst in self.profiles:
            inst_org = org_vm.tovalues(inst['RegisteredOrganization'])
            inst_name = inst['RegisteredName']
            inst_version = inst['RegisteredVersion']

            # pylint: disable=too-many-boolean-expressions
            if (registered_org is None or registered_org == inst_org) and \
                    (registered_name is None or registered_name == inst_name) \
                    and \
                    (registered_version is None or
                     registered_version == inst_version):
                rtn.append(inst)
        return rtn

    def get_central_instances(self, profile_path, central_class=None,
                              scoping_class=None, scoping_path=None):
        """
        Determine the central instances for a management profile, and return
        their instance paths as a list of :class:`~pywbem.CIMInstanceName`
        objects.

        This method supports the following profile advertisement methodologies
        (see :term:`DSP1033`), and attempts them in this order:

        * GetCentralInstances methodology (new in :term:`DSP1033` 1.1)
        * Central class methodology
        * Scoping class methodology

        Use of the scoping class methodology requires specifying the central
        class, scoping class and scoping path defined by the profile. If any
        of them is `None`, this method will attempt only the GetCentralInstances
        and central class methodologies, but not the scoping class methodology.
        If using these two methodologies does not result in any central
        instances, and the scoping class methodology cannot be used, an
        exception is raised.

        The scoping path is a directed traversal path from the central
        instances to the scoping instances. Its first list item is always
        the association class name of the traversal hop starting at the
        central instances. For each further traversal hop, the list contains
        two more items: The class name of the near end of that hop, and
        the class name of the traversed association.
        As a result, the class names of the central instances and scoping
        instances are not part of the list.

        Example for a 1-hop traversal:

        * central class: ``"CIM_Fan"``
        * scoping path: ``["CIM_SystemDevice"]``
        * scoping class: ``"CIM_ComputerSystem"``

        Example for a 2-hop traversal:

        * central class: ``"CIM_Sensor"``
        * scoping path: ``["CIM_AssociatedSensor", "CIM_Fan", \
                           "CIM_SystemDevice"]``
        * scoping class: ``"CIM_ComputerSystem"``

        Parameters:

          profile_path (:class:`~pywbem.CIMInstanceName`):
            Instance path of CIM_RegisteredProfile instance representing the
            management profile.

          central_class (:term:`string`):
            Class name of central class defined by the management profile.

            Will be ignored, unless the profile is a component profile and its
            implementation supports only the scoping class methodology.
            `None` will cause the scoping class methodology not to be attempted.

          scoping_class (:term:`string`):
            Class name of scoping class defined by the management profile.

            Will be ignored, unless the profile is a component profile and its
            implementation supports only the scoping class methodology.
            `None` will cause the scoping class methodology not to be attempted.

          scoping_path (list of :term:`string`):
            Scoping path defined by the management profile.

            Will be ignored, unless the profile is a component profile and its
            implementation supports only the scoping class methodology.
            `None` will cause the scoping class methodology not to be attempted.

        Returns:

          List of :class:`~pywbem.CIMInstanceName` objects representing the
          instance paths of the central instances of the management profile.

        Raises:

            Exceptions raised by :class:`~pywbem.WBEMConnection`.
            ValueError: Various errors in scoping path traversal.
            TypeError: profile_path must be a CIMInstanceName.
        """
        if not isinstance(profile_path, CIMInstanceName):
            raise TypeError("profile_path must be a CIMInstanceName, but is "
                            "a %s" % type(profile_path))

        # Try GetCentralInstances() method:
        try:
            (ret_val, out_params) = self._conn.InvokeMethod(
                MethodName="GetCentralInstances",
                ObjectName=profile_path)
        except CIMError as exc:
            if exc.status_code in (CIM_ERR_METHOD_NOT_AVAILABLE,
                                   CIM_ERR_NOT_SUPPORTED):
                # Method is not implemented.
                # CIM_ERR_NOT_SUPPORTED is not an official status code for this
                # situation, but is used by some implementations.
                pass  # try next approach
            else:
                raise
        else:
            if ret_val != 0:
                raise ValueError("GetCentralInstances() implemented but "
                                 "failed with rc=%s" % ret_val)
            return out_params['CentralInstances']

        # Try central methodology
        ci_paths = self._conn.AssociatorNames(
            ObjectName=profile_path,
            AssocClass="CIM_ElementConformsToProfile",
            ResultRole="ManagedElement")
        if len(ci_paths) > 0:
            return ci_paths

        # Try scoping methodology
        if central_class is None or \
           scoping_class is None or \
           scoping_path is None:
            raise ValueError("No central instances found after applying "
                             "GetCentralInstances and central class "
                             "methodologies, and parameters for scoping "
                             "class methodology were not specified")

        # Go up one level on the profile side
        referencing_profile_paths = self._conn.AssociatorNames(
            ObjectName=profile_path,
            AssocClass="CIM_ReferencedProfile",
            ResultRole="Dependent")
        if len(referencing_profile_paths) == 0:
            raise ValueError("No referencing profile found")
        elif len(referencing_profile_paths) > 1:
            raise ValueError("More than one referencing profile found")

        # Traverse to the resource side (remember that scoping instances are
        # the central instances at the next upper level).
        # Do this recursively, if needed.
        if len(scoping_path) >= 3:
            upper_central_class = scoping_path[1]
            upper_scoping_path = scoping_path[2:-1]
        else:
            upper_central_class = None
            upper_scoping_path = None
        scoping_inst_paths = self.get_central_instances(
            referencing_profile_paths[0],
            upper_central_class, scoping_class, upper_scoping_path)
        if len(scoping_inst_paths) == 0:
            raise ValueError("No scoping instances found")

        # Go down one level on the resource side (using the last
        # entry in the scoping path as the association to traverse)
        total_ci_paths = []
        assoc_class = scoping_path[-1]
        for ip in scoping_inst_paths:
            ci_paths = self._conn.AssociatorNames(
                ObjectName=ip,
                AssocClass=assoc_class,
                ResultClass=central_class)
            if len(ci_paths) == 0:
                # At least one central instance for each scoping instance
                raise ValueError("No central instances found traversing down "
                                 "across %s to %s" %
                                 (assoc_class, central_class))
            total_ci_paths.extend(ci_paths)

        return total_ci_paths

    def _determine_interop_ns(self):
        """
        Determine the name of the Interop namespace of the WBEM server, by
        trying to communicate with it on a number of possible Interop
        namespace names, that are defined in the :attr:`INTEROP_NAMESPACES`
        class variable.

        If the Interop namespace could be determined, this method sets the
        :attr:`interop_ns` property of this object to that namespace and
        returns.
        Otherwise, it raises an exception.

        Raises:

            Exceptions raised by :class:`~pywbem.WBEMConnection`.
            CIMError: CIM_ERR_NOT_FOUND, Interop namespace could not be
              determined.
        """
        test_classname = 'CIM_Namespace'
        interop_ns = None
        for ns in self.INTEROP_NAMESPACES:
            try:
                inst_paths = self._conn.EnumerateInstanceNames(test_classname,
                                                               namespace=ns)
            except CIMError as exc:
                if exc.status_code == CIM_ERR_INVALID_NAMESPACE:
                    # Current namespace does not exist.
                    continue
                elif exc.status_code in (CIM_ERR_INVALID_CLASS,
                                         CIM_ERR_NOT_FOUND):
                    # Class is not implemented, but current namespace exists.
                    interop_ns = ns
                    break
                else:
                    # Some other error happened.
                    raise
            else:
                # Namespace class is implemented in the current namespace.
                # Use the returned namespace name, if possible.
                ns_names = [p.keybindings['name'] for p in inst_paths]
                ns_dict = NocaseDict(list(zip(ns_names, ns_names)))
                try:
                    interop_ns = ns_dict[ns]
                except KeyError:
                    interop_ns = ns
                break
        if interop_ns is None:
            # Exhausted the possible namespaces
            raise CIMError(CIM_ERR_NOT_FOUND,
                           "Interop namespace could not be determined "
                           "(tried %s)" % self.INTEROP_NAMESPACES)
        self._interop_ns = interop_ns

    def _validate_interop_ns(self, interop_ns):
        """
        Validate whether the specified Interop namespace exists in the WBEM
        server, by communicating with it.

        If the specified Interop namespace exists, this method sets the
        :attr:`interop_ns` property of this object to that namespace and
        returns.
        Otherwise, it raises an exception.

        Parameters:

          interop_ns (:term:`string`):
            Name of the Interop namespace to be validated.

        Raises:

            Exceptions raised by :class:`~pywbem.WBEMConnection`.
        """
        test_classname = 'CIM_Namespace'
        try:
            self._conn.EnumerateInstanceNames(test_classname,
                                              namespace=interop_ns)
        except CIMError as exc:
            # We tolerate it if the WBEM server does not implement this class,
            # as long as it does not return CIM_ERR_INVALID_NAMESPACE.
            if exc.status_code in (CIM_ERR_INVALID_CLASS,
                                   CIM_ERR_NOT_FOUND):
                pass
            else:
                raise
        self._interop_ns = interop_ns

    def _determine_namespaces(self):
        """
        Determine the names of all namespaces of the WBEM server, by
        communicating with it and enumerating the instances of a number of
        possible CIM classes that typically represent CIM namespaces. Their
        class names are defined in the :attr:`NAMESPACE_CLASSNAMES`
        class variable.

        If the namespaces could be determined, this method sets the
        :attr:`namespace_classname` property of this object to the class name
        that was found to work, the :attr:`namespaces` property to these
        namespaces, and returns.
        Otherwise, it raises an exception.

        Raises:

            Exceptions raised by :class:`~pywbem.WBEMConnection`.
            CIMError: CIM_ERR_NOT_FOUND, Interop namespace could not be
              determined.
            CIMError: CIM_ERR_NOT_FOUND, Namespace class could not be
              determined.
        """
        ns_insts = None
        ns_classname = None
        for classname in self.NAMESPACE_CLASSNAMES:
            try:
                ns_insts = self._conn.EnumerateInstances(
                    classname, namespace=self.interop_ns)
            except CIMError as exc:
                if exc.status_code in (CIM_ERR_INVALID_CLASS,
                                       CIM_ERR_NOT_FOUND):
                    # Class is not implemented, try next one.
                    continue
                else:
                    # Some other error.
                    raise
            else:
                # Found a namespace class that is implemented.
                ns_classname = classname
                break
        if ns_insts is None:
            # Exhausted the possible class names
            raise CIMError(CIM_ERR_NOT_FOUND,
                           "Namespace class could not be determined "
                           "(tried %s)" % self.NAMESPACE_CLASSNAMES)
        self._namespace_classname = ns_classname
        self._namespaces = [inst['Name'] for inst in ns_insts]

    def _determine_brand(self):
        """
        Determine the brand of the WBEM server (e.g. OpenPegasus, SFCB, ...)
        and its version, by communicating with it and retrieving the
        CIM_ObjectManager instance.

        On success, this method sets the :attr:`brand` and :attr:`version`
        properties of this object and returns.
        Otherwise, it raises an exception.

        Raises:

            Exceptions raised by :class:`~pywbem.WBEMConnection`.
            CIMError: CIM_ERR_NOT_FOUND, Interop namespace could not be
              determined.
            CIMError: CIM_ERR_NOT_FOUND, Unexpected number of
              CIM_ObjectManager instances.
        """
        cimom_insts = self._conn.EnumerateInstances(
            "CIM_ObjectManager", namespace=self.interop_ns)
        if len(cimom_insts) != 1:
            raise CIMError(CIM_ERR_NOT_FOUND,
                           "Unexpected number of CIM_ObjectManager "
                           "instances: %s " %
                           [i['ElementName'] for i in cimom_insts])
        cimom_inst = cimom_insts[0]
        element_name = cimom_inst['ElementName']
        if element_name is not None:
            element_word = element_name.split(' ')[0]
        else:
            element_word = "unknown"
        if element_word in ("Pegasus", "OpenPegasus"):
            brand = 'OpenPegasus'
            # Description = "Pegasus OpenPegasus Version 2.12.0"
            m = re.match(r'.+ *Version *([^ ]+)', cimom_inst['Description'])
            if m:
                version = m.group(1)
            else:
                version = None
        elif element_word == "SFCB":
            brand = 'SFCB'
            # TODO: Figure out how to get version of SFCB
            version = None
        else:
            brand = element_word
            version = None
        self._brand = brand
        self._version = version

    def _determine_profiles(self):
        """
        Determine the WBEM management profiles advertised by the WBEM server,
        by communicating with it and enumerating the instances of
        CIM_RegisteredProfile.

        If the profiles could be determined, this method sets the
        :attr:`profiles` property of this object to the list of
        CIM_RegisteredProfile instances (as :class:`~pywbem.CIMInstance`
        objects), and returns.
        Otherwise, it raises an exception.

        Raises:

            Exceptions raised by :class:`~pywbem.WBEMConnection`.
            CIMError: CIM_ERR_NOT_FOUND, Interop namespace could not be
              determined.
        """
        mp_insts = self._conn.EnumerateInstances("CIM_RegisteredProfile",
                                                 namespace=self.interop_ns)
        self._profiles = mp_insts


class ValueMapping(object):
    """
    A utility class that translates the values of a corresponding integer-typed
    CIM element (property, method, parameter) that is qualified with the
    ValueMap and Values qualifiers, from the element value space into into its
    Values qualifier space.

    This is done by retrieving the CIM class definition defining the CIM
    element in question, and by inspecting its ValueMap and Values qualifiers.

    The actual translation of the values is performed by the
    :meth:`~pywbem.ValueMapping.tovalues` method.

    Instances of this class should be created through one of the factory class
    methods: :meth:`~pywbem.ValueMapping.for_property`,
    :meth:`~pywbem.ValueMapping.for_method`, or
    :meth:`~pywbem.ValueMapping.for_parameter`.

    Value ranges (``"2..4"``) and the indicator for unclaimed values (``".."``)
    in the ValueMap qualifier are supported.

    Example: Given the following definition of a property in MOF:

    ::

        [ValueMap{ "0", "2..4", "..6", "7..", "9", ".." },
         Values{ "zero", "two-four", "five-six", "seven-eight", "nine", \
             "unclaimed"}]
        uint16 MyProp;

    The following code will create a value mapping for this property and will
    print a few integer values and their Values strings:

    ::

        vm = pywbem.ValueMapping.for_property(server, namespace, classname,\
                "MyProp")
        for value in range(0, 12):
            print("value: %s, Values string: %r" % (value, vm.tovalues(value))

        Results:
        value: 0, Values string: 'zero'
        value: 1, Values string: 'unclaimed'
        value: 2, Values string: 'two-four'
        value: 3, Values string: 'two-four'
        value: 4, Values string: 'two-four'
        value: 5, Values string: 'five-six'
        value: 6, Values string: 'five-six'
        value: 7, Values string: 'seven-eight'
        value: 8, Values string: 'seven-eight'
        value: 9, Values string: 'nine'
        value: 10, Values string: 'unclaimed'
        value: 11, Values string: 'unclaimed'
    """

    def __init__(self):
        self._element_obj = None
        self._single_dict = {}  # for single values; elem_val: values_str)
        self._range_tuple_list = []  # for value ranges; tuple(lo,hi,values_str)
        self._unclaimed = None  # value of the unclaimed indicator '..'

    @classmethod
    def for_property(cls, server, namespace, classname, propname):
        """
        Factory method that returns a new :class:`~pywbem.ValueMapping`
        instance corresponding to a CIM property.

        If a Values qualifier is defined but no ValueMap qualifier, a default
        of 0-based consecutive numbers is applied (that is the default defined
        in :term:`DSP0004`).

        Parameters:

          server (:class:`~pywbem.WBEMServer`):
            The WBEM server containing the namespace.

          namespace (:term:`string`):
            Name of the CIM namespace containing the class.

          classname (:term:`string`):
            Name of the CIM class exposing the property. The property can be
            defined in that class or inherited into that class.

          propname (:term:`string`):
            Name of the CIM property that defines the Values / ValueMap
            qualifiers.

        Returns:

            The new :class:`~pywbem.ValueMapping` instance.

        Raises:

            Exceptions raised by :class:`~pywbem.WBEMConnection`.
            ValueError: No Values qualifier defined.
            TypeError: The property is not integer-typed.
        """
        class_obj = server.conn.GetClass(ClassName=classname,
                                         namespace=namespace,
                                         LocalOnly=False,
                                         IncludeQualifiers=True)
        property_obj = class_obj.properties[propname]
        return cls._create_for_element(property_obj)

    @classmethod
    def for_method(cls, server, namespace, classname, methodname):
        """
        Factory method that returns a new :class:`~pywbem.ValueMapping`
        instance corresponding to a CIM method.

        If a Values qualifier is defined but no ValueMap qualifier, a default
        of 0-based consecutive numbers is applied (that is the default defined
        in :term:`DSP0004`).

        Parameters:

          server (:class:`~pywbem.WBEMServer`):
            The WBEM server containing the namespace.

          namespace (:term:`string`):
            Name of the CIM namespace containing the class.

          classname (:term:`string`):
            Name of the CIM class exposing the method. The method can be
            defined in that class or inherited into that class.

          methodname (:term:`string`):
            Name of the CIM method that defines the Values / ValueMap
            qualifiers.

        Returns:

            The new :class:`~pywbem.ValueMapping` instance.

        Raises:

            Exceptions raised by :class:`~pywbem.WBEMConnection`.
            ValueError: No Values qualifier defined.
            TypeError: The method is not integer-typed.
        """
        class_obj = server.conn.GetClass(ClassName=classname,
                                         namespace=namespace,
                                         LocalOnly=False,
                                         IncludeQualifiers=True)
        method_obj = class_obj.methods[methodname]
        return cls._create_for_element(method_obj)

    @classmethod
    def for_parameter(cls, server, namespace, classname, methodname,
                      parametername):
        """
        Factory method that returns a new :class:`~pywbem.ValueMapping`
        instance corresponding to a CIM parameter.

        If a Values qualifier is defined but no ValueMap qualifier, a default
        of 0-based consecutive numbers is applied (that is the default defined
        in :term:`DSP0004`).

        Parameters:

          server (:class:`~pywbem.WBEMServer`):
            The WBEM server containing the namespace.

          namespace (:term:`string`):
            Name of the CIM namespace containing the class.

          classname (:term:`string`):
            Name of the CIM class exposing the method. The method can be
            defined in that class or inherited into that class.

          methodname (:term:`string`):
            Name of the CIM method that has the parameter.

          parametername (:term:`string`):
            Name of the CIM parameter that defines the Values / ValueMap
            qualifiers.

        Returns:

            The new :class:`~pywbem.ValueMapping` instance.

        Raises:

            Exceptions raised by :class:`~pywbem.WBEMConnection`.
            ValueError: No Values qualifier defined.
            TypeError: The parameter is not integer-typed.
        """
        class_obj = server.conn.GetClass(ClassName=classname,
                                         namespace=namespace,
                                         LocalOnly=False,
                                         IncludeQualifiers=True)
        method_obj = class_obj.methods[methodname]
        parameter_obj = method_obj.parameters[parametername]
        return cls._create_for_element(parameter_obj)

    @classmethod
    def _values_tuple(cls, i, valuemap_list, values_list, cimtype):
        """
        Return a tuple for the value range at position i, with these items:

        * lo - low value of the range
        * hi - high value of the range (can be equal to lo)
        * values - Value of Values qualifier for this position

        Parameters:

          i (integer): position into valuemap_list and values_list

          valuemap_list (list of strings): ValueMap qualifier value

          values_list (list of strings): Values qualifier value

          cimtype (type): CIM type of the CIM element

        Raises:

            ValueError: Invalid ValueMap entry.
        """
        values_str = values_list[i]
        valuemap_str = valuemap_list[i]
        try:
            valuemap_int = int(valuemap_str)
            return (valuemap_int, valuemap_int, values_str)
        except ValueError:
            m = re.match(r'^([0-9]*)\.\.([0-9]*)$', valuemap_str)
            if m is None:
                raise ValueError("Invalid ValueMap entry: %r" % valuemap_str)
            lo = m.group(1)
            if lo == '':
                if i == 0:
                    lo = 0
                    # TODO: Change to min(cimtype) once issue #268 is solved.
                else:
                    _, previous_hi, _ = cls._values_tuple(
                        i - 1, valuemap_list, values_list, cimtype)
                    lo = previous_hi + 1
            else:
                lo = int(lo)
            hi = m.group(2)
            if hi == '':
                if i == len(valuemap_list) - 1:
                    hi = 32767
                    # TODO: Change to max(cimtype) once issue #268 is solved.
                else:
                    next_lo, _, _ = cls._values_tuple(
                        i + 1, valuemap_list, values_list, cimtype)
                    hi = next_lo - 1
            else:
                hi = int(hi)
            return (lo, hi, values_str)

    @classmethod
    def _create_for_element(cls, element_obj):
        # pylint: disable=line-too-long
        """
        Return a new :class:`~pywbem.ValueMapping` instance for the specified
        CIM element.

        If a Values qualifier is defined but no ValueMap qualifier, a default
        of 0-based consecutive numbers is applied (that is the default defined
        in :term:`DSP0004`).

        Parameters:

          element_obj (:class:`~pywbem.CIMProperty`, :class:`~pywbem.CIMMethod`, or :class:`~pywbem.CIMParameter`):
            The CIM element on which the qualifiers are defined.

        Returns:

            The created :class:`~pywbem.ValueMapping` instance for the specified
            CIM element.

        Raises:

            ValueError: No Values qualifier defined.
            ValueError: Invalid ValueMap entry.
            TypeError: The CIM element is not integer-typed.
        """  # noqa: E501

        # pylint: disable=protected-access

        typename = element_obj.type
        # TODO: We should make type_from_name() and cimtype() part of the API
        cimtype = type_from_name(typename)

        if not issubclass(cimtype, CIMInt):
            raise TypeError("The CIM element is not integer-typed: %s" %
                            typename)

        vm = ValueMapping()
        vm._element_obj = element_obj

        values_qual = element_obj.qualifiers.get('Values', None)
        if values_qual is None:
            # DSP0004 defines no default for a missing Values qualifier
            raise ValueError("No Values qualifier defined")
        values_list = values_qual.value

        valuemap_qual = element_obj.qualifiers.get('ValueMap', None)
        if valuemap_qual is None:
            # DSP0004 defines a default of consecutive index numbers
            vm._single_dict = dict(zip(range(0, len(values_list)), values_list))
            vm._range_tuple_list = []
            vm._unclaimed = None
        else:
            vm._single_dict = {}
            vm._range_tuple_list = []
            vm._unclaimed = None
            valuemap_list = valuemap_qual.value
            for i, valuemap_str in enumerate(valuemap_list):
                values_str = values_list[i]
                if valuemap_str == '..':
                    vm._unclaimed = values_str
                else:
                    lo, hi, values_str = cls._values_tuple(
                        i, valuemap_list, values_list, cimtype)
                    if lo == hi:
                        # single value
                        vm._single_dict[lo] = values_str
                    else:
                        # value range
                        vm._range_tuple_list.append((lo, hi, values_str))

        return vm

    @property
    def element(self):
        """
        Return the corresponding CIM element of this instance, as a CIM object
        (:class:`~pywbem.CIMProperty`, :class:`~pywbem.CIMMethod`, or
        :class:`~pywbem.CIMParameter`).
        """
        return self._element_obj

    def tovalues(self, element_value):
        """
        Return the Values string for an element value, based on the ValueMap /
        Values qualifiers of the corresponding CIM element.

        Parameters:

          element_value (:term:`integer` or :class:`~pywbem.CIMInt`):
            The value of the CIM element.

        Returns:

          :term:`string`:
            The Values string for the element value.

        Raises:

            ValueError: Element value outside of the set defined by ValueMap.
            ValueError: No Values qualifier defined.
            ValueError: Invalid ValueMap entry.
            TypeError: The CIM element is not integer-typed.
            TypeError: Element value is not an integer type.
        """

        if not isinstance(element_value, (six.integer_types, CIMInt)):
            raise TypeError("Element value is not an integer type: %s" %
                            type(element_value))

        # try single value
        try:
            return self._single_dict[element_value]
        except KeyError:
            pass

        # try value ranges
        for range_tuple in self._range_tuple_list:
            lo, hi, values_str = range_tuple
            if lo <= element_value <= hi:
                return values_str

        # try catch-all '..'
        if self._unclaimed is not None:
            return self._unclaimed

        raise ValueError("Element value outside of the set defined by "
                         "ValueMap: %r" % element_value)
